#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "../include/nicslu.h"

int n = 0;
double *ax = NULL;
int *ai = NULL;
int *ap = NULL;

int ReadFileToCSR(const char *file);

main(void)
{
	int ret, nnz, i;
	SNicsLU *nicslu;
	double *x, *b, err;
	x = NULL;

	nicslu = (SNicsLU *)malloc(sizeof(SNicsLU));
	NicsLU_Initialize(nicslu);

	ret = ReadFileToCSR("ASIC_100k.mtx");
	if (ret != 0) goto EXIT;

	nnz = ap[n];
	x = (double *)malloc(sizeof(double)*(n+n));
	b = x + n;
	for (i=0; i<n+n; ++i) x[i] = 1.;

	NicsLU_CreateMatrix(nicslu, n, nnz, ax, (uint__t *)ai, (uint__t *)ap);

	NicsLU_Analyze(nicslu);
	printf("analysis time: %.8g\n", nicslu->stat[0]);

	NicsLU_Factorize(nicslu);
	printf("factorization time: %.8g\n", nicslu->stat[1]);

	NicsLU_ReFactorize(nicslu, ax);
	printf("re-factorization time: %.8g\n", nicslu->stat[2]);

	NicsLU_Solve(nicslu, x);
	printf("substitution time: %.8g\n", nicslu->stat[3]);

	NicsLU_Residual(n, ax, ai, ap, x, b, &err, 1, 0);
	printf("Ax-b (1-norm): %.8g\n", err);
	NicsLU_Residual(n, ax, ai, ap, x, b, &err, 2, 0);
	printf("Ax-b (2-norm): %.8g\n", err);
	NicsLU_Residual(n, ax, ai, ap, x, b, &err, 0, 0);
	printf("Ax-b (infinite-norm): %.8g\n", err);

	printf("NNZ(L+U-I): %ld\n", nicslu->lu_nnz);

	NicsLU_Flops(nicslu, NULL);
	NicsLU_Throughput(nicslu, NULL);
	NicsLU_ConditionNumber(nicslu, NULL);
	printf("flops: %.8g\n", nicslu->stat[5]);
	printf("throughput (bytes): %.8g\n", nicslu->stat[12]);
	printf("condition number: %.8g\n", nicslu->stat[6]);
	
EXIT:
	NicsLU_Destroy(nicslu);
	free(ax);
	free(ai);
	free(ap);
	free(nicslu);
	free(x);
}
