#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "../include/nicslu.h"

extern int n;
extern double *ax;
extern int *ai;
extern int *ap;

#define MM_MAX_LINE_LENGTH 1025

int mm_read_mtx_crd_size(FILE *f, int *M, int *N, int *nz )
{
    char line[MM_MAX_LINE_LENGTH];
    int num_items_read;

    /* set return null parameter values, in case we exit with errors */
    *M = *N = *nz = 0;

    /* now continue scanning until you reach the end-of-comments */
    do 
    {
        if (fgets(line,MM_MAX_LINE_LENGTH,f) == NULL) 
            return 1;
    }while (line[0] == '%');

    /* line[] is either blank or has M,N, nz */
    if (sscanf(line, "%d %d %d", M, N, nz) == 3)
        return 0;
        
    else
    do
    { 
        num_items_read = fscanf(f, "%d %d %d", M, N, nz); 
        if (num_items_read == EOF) return 1;
    }
    while (num_items_read != 3);

    return 0;
}

int ReadFileToCSR(const char *file)
{
	FILE *fp;
	int m, nnz;
	int ret, i, j;
	int *J;
	int precol, num, col;

	fp = NULL;
	J = NULL;

	fp = fopen(file, "r");
	if (fp == NULL)
	{
		printf("cannot open file.\n");
		goto FAIL;
	}

	ret = mm_read_mtx_crd_size(fp, &m, &n, &nnz);
	if (ret != 0)
	{
		printf("an error occurs when reading the file.\n");
		goto FAIL;
	}
	if (m != n)
	{
		printf("this application does not support non-square matrix.\n");
		goto FAIL;
	}

	ai = (int *)malloc(nnz*sizeof(int));
    J = (int *)malloc(nnz*sizeof(int));
    ax = (double *)malloc(nnz*sizeof(double));

	for (i=0; i<nnz; ++i)
	{
		fscanf(fp, "%d %d %lf", &ai[i], &J[i], &ax[i]);
		--ai[i];
		--J[i];
	}

	ap = (int *)malloc(sizeof(int)*(1+n));
	*ap = 0;
	precol = 0;
	num = 0;

	for (i=0; i<nnz; ++i)
	{
		col = J[i];
		if (precol == col)
		{
			++num;
		}
		else
		{
			num += ap[precol];
			for (j=precol+1; j<=col; j++)
			{
				ap[j] = num;
			}
			precol = col;
			num = 1;
		}
	}
	num += ap[col];
	for (i=col+1; i<=n; i++)
	{
		ap[i] = num;
	}

	ret = NicsLU_Transpose(n, nnz, (real__t *)ax, (uint__t *)ai, (uint__t *)ap);
	if (ret < 0)
	{
		printf("an error occurs when transposing the matrix.\n");
		goto FAIL;
	}

	fclose(fp);
	free(J);
	return 0;

FAIL:
	if (fp != NULL) fclose(fp);
	free(J);
	free(ax);
	free(ai);
	free(ap);
	return -1;
}
