/*evaluate the ideal maximum performance*/
/*last modified: apr 14, 2013*/
/*author: Chen, Xiaoming*/

#include "nicslu.h"
#include "nicslu_internal.h"
#include "math.h"

extern size_t g_si, g_sd, g_sp;

int__t NicsLU_Performance(SNicsLU *nicslu, real__t *perf, real__t *tperf, uint__t thread)
{
	int__t rt;
	real__t flops;
	real__t pflops;
	uint__t n, i, j, k, minid;
	real__t *end, *thfi;
	size_t *up;
	uint__t *ulen, *llen;
	uint__t *lrow;
	uint__t ll, ul;
	void *lu;
	real__t ed, min, tsync;

	if (NULL == nicslu || 0 == thread)
	{
		return NICSLU_ARGUMENT_ERROR;
	}

	if (perf != NULL) *perf = -1.;
	if (tperf != NULL) *tperf = -1.;
	tsync = nicslu->cfgf[3];

	if (!nicslu->flag[2])
	{
		return NICSLU_MATRIX_NOT_FACTORIZED;
	}

	rt = NicsLU_Flops(nicslu, &flops);
	if (rt < 0) return rt;

	n = nicslu->n;
	up = nicslu->up;
	llen = nicslu->llen;
	ulen = nicslu->ulen;
	lu = nicslu->lu_array;

	end = (real__t *)nicslu->workspace;
	thfi = end + n;
	memset(end, 0, sizeof(real__t)*(n<<1));

	/*perfect performance*/
	pflops = 0.;
	for (i=0; i<n; ++i)
	{
		ll = llen[i];
		ul = ulen[i];

		lrow = (uint__t *)(((byte__t *)lu) + up[i] + ul*g_sp);

		ed = 0.;
		for (j=0; j<ll; ++j)
		{
			k = lrow[j];
			ed = MAX(ed, end[k]);
			ed += (ulen[k]<<1);
		}
		ed += ul;
		end[i] = ed;

		if (ed > pflops)
		{
			pflops = ed;
		}
	}

	if (pflops == 0.) nicslu->stat[18] = n;
	else nicslu->stat[18] = flops / pflops;
	if (perf != NULL) *perf = nicslu->stat[18];

	//thread-specific performance
	pflops = 0.;
	for (i=0; i<n; ++i)
	{
		min = DBL_MAX;
		for (j=0; j<thread; ++j)
		{
			if (thfi[j] < min)
			{
				min = thfi[j];
				minid = j;
			}
		}

		ll = llen[i];
		ul = ulen[i];

		lrow = (uint__t *)(((byte__t *)lu) + up[i] + ul*g_sp);

		ed = min;
		for (j=0; j<ll; ++j)
		{
			k = lrow[j];
			ed = MAX(ed, end[k]);
			ed += (ulen[k]<<1);
			ed += tsync;
		}
		ed += ul;
		ed += tsync;
		end[i] = ed;
		thfi[minid] = ed;

		if (ed > pflops)
		{
			pflops = ed;
		}
	}

	if (pflops == 0.) nicslu->stat[17] = n;
	else nicslu->stat[17] = flops / pflops;
	if (tperf != NULL) *tperf = nicslu->stat[17];

	return NICS_OK;
}
