/*create and destroy the scheduler*/
/*last modified: apr 14, 2013*/
/*author: Chen, Xiaoming*/

#include "nicslu.h"
#include "nicslu_internal.h"
#include "timer_c.h"

int__t NicsLU_CreateScheduler(SNicsLU *nicslu/*, real__t *pred, uint__t *cores*/)
{
	uint__t n;
	size_t size;
	void *ptr;
	int__t rt;
	bool__t a, b;
	real__t nnz;
	
	if (NULL == nicslu/* || NULL == pred || NULL == cores*/)
	{
		return NICSLU_ARGUMENT_ERROR;
	}
/*	if (pred != NULL) *pred = -1.;
	if (cores != NULL) *cores = 0;*/
	if (!nicslu->flag[1])
	{
		return NICSLU_MATRIX_NOT_ANALYZED;
	}

	TimerStart((STimer *)(nicslu->timer));
	NicsLU_DestroyScheduler(nicslu);

	n = nicslu->n;

	/*wkld_est & len_est*/
	size = sizeof(size_t)*n*2;
	ptr = malloc(size);
	if (NULL == ptr)
	{
		return NICSLU_MEMORY_OVERFLOW;
	}
	nicslu->wkld_est = (size_t *)ptr;
	ptr = ((size_t *)ptr) + n;
	nicslu->len_est = (size_t *)ptr;

	/*row state*/
	size = sizeof(byte__t)*n;
	ptr = malloc(size);
	if (NULL == ptr)
	{
		return NICSLU_MEMORY_OVERFLOW;
	}
	nicslu->row_state = (byte__t *)ptr;

	/*scheduler*/
	size = sizeof(uint__t) * (4*n+2);
	ptr = malloc(size);
	if (NULL == ptr)
	{
		return NICSLU_MEMORY_OVERFLOW;
	}
	nicslu->aeg_data = (uint__t *)ptr;
	ptr = ((uint__t *)ptr) + n;
	nicslu->aeg_header = (uint__t *)ptr;
	ptr = ((uint__t *)ptr) + n + 1;
	nicslu->aeg_refact_data = (uint__t *)ptr;
	ptr = ((uint__t *)ptr) + n;
	nicslu->aeg_refact_header = (uint__t *)ptr;

	/*lu_array2*/
	size = sizeof(void *) * n;
	ptr = malloc(size);
	if (NULL == ptr)
	{
		return NICSLU_MEMORY_OVERFLOW;
	}
	memset(ptr, 0, size);
	nicslu->lu_array2 = (void **)ptr;

	/*calculate the scheduler*/
	rt = _I_NicsLU_CreateETree(nicslu);
	if (rt < 0) return rt;

	rt = _I_NicsLU_StaticSymbolicFactorize(nicslu);
	if (rt < 0) return rt;

	/*finish*/
	nicslu->flag[4] = TRUE;
	
	nnz = nicslu->stat[11];
	a = ((nnz/(real__t)nicslu->nnz) >= 1.99999999);
	b = ((nicslu->stat[10]/nnz) >= 50.);

	TimerStop((STimer *)(nicslu->timer));
	nicslu->stat[4] = TimerGetRuntime((STimer *)(nicslu->timer));

	rt = (!a && !b);
	nicslu->stat[13] = rt;
	return rt;
}

int__t NicsLU_DestroyScheduler(SNicsLU *nicslu)
{
	uint__t i, n;
	void **ppv;

	if (NULL == nicslu)
	{
		return NICSLU_ARGUMENT_ERROR;
	}
	n = nicslu->n;

	if (nicslu->wkld_est != NULL)
	{
		free(nicslu->wkld_est);
		nicslu->wkld_est = NULL;
	}
	nicslu->len_est = NULL;
	if (nicslu->row_state != NULL)
	{
		free(nicslu->row_state);
		nicslu->row_state = NULL;
	}
	if (nicslu->aeg_data != NULL)
	{
		free(nicslu->aeg_data);
		nicslu->aeg_data = NULL;
	}
	nicslu->aeg_header = NULL;
	nicslu->aeg_refact_data = NULL;
	nicslu->aeg_refact_header = NULL;

	if (nicslu->lu_array2 != NULL)
	{
		ppv = nicslu->lu_array2;
		for (i=0; i<n; ++i)
		{
			if (ppv[i] != NULL)
			{
				free(ppv[i]);
			}
		}
		free(ppv);
		nicslu->lu_array2 = NULL;
	}

	nicslu->flag[4] = FALSE;
	return NICS_OK;
}

