/*create and destroy threads*/
/*last modified: apr 14, 2013*/
/*author: Chen, Xiaoming*/

#include "nicslu.h"
#include "nicslu_internal.h"
#include "thread.h"
#include "system.h"

int__t NicsLU_CreateThreads(SNicsLU *nicslu, uint__t threads, bool__t check)
{
	uint__t i, n;
	uint__t cores;
	void *ptr;
	size_t size;
	SNicsLUThreadArg *tharg;

	if (NULL == nicslu)
	{
		return NICSLU_ARGUMENT_ERROR;
	}
	if (!nicslu->flag[0])
	{
		return NICSLU_MATRIX_NOT_INITIALIZED;
	}

	n = nicslu->n;

	/*check threads*/
	if (check)
	{
		cores = nicslu->cfgi[31];
		if (threads > cores)
		{
			threads = cores;
		}
	}
	if (threads == 0)
	{
		return NICSLU_ARGUMENT_ERROR;
	}
	if (threads == 1)
	{
		return NICSLU_SINGLE_THREAD;
	}

	/*exit the previous threads first*/
	NicsLU_DestroyThreads(nicslu);
	nicslu->cfgi[5] = threads;
	nicslu->cfgi[7] = threads;

	/*thread_id*/
	size = sizeof(thread_id__t)*threads;
	ptr = malloc(size);
	if (NULL == ptr)
	{
		return NICSLU_MEMORY_OVERFLOW;
	}
	memset(ptr, 0, size);
	nicslu->thread_id = ptr;

	/*thread arg*/
	size = sizeof(SNicsLUThreadArg)*threads;
	ptr = malloc(size);
	if (NULL == ptr)
	{
		return NICSLU_MEMORY_OVERFLOW;
	}
	nicslu->thread_arg = ptr;
	for (i=0; i<threads; ++i)
	{
		tharg = &(((SNicsLUThreadArg *)ptr)[i]);
		tharg->nicslu = nicslu;
		tharg->no = i;
	}

	/*thread work*/
	nicslu->thread_work = NICSLU_WORK_NONE;

	/*thread active & thread finish*/
	size = sizeof(bool__t)*(threads<<1);
	ptr = malloc(size);
	if (NULL == ptr)
	{
		return NICSLU_MEMORY_OVERFLOW;
	}
	memset(ptr, 0, size);
	nicslu->thread_finish = (bool__t *)ptr;
	ptr = ((bool__t *)ptr) + threads;
	nicslu->thread_active = (bool__t *)ptr;

	/*cluster start and end*/
	size = sizeof(uint__t)*(threads<<1);
	ptr = malloc(size);
	if (NULL == ptr)
	{
		return NICSLU_MEMORY_OVERFLOW;
	}
	nicslu->cluster_start = (uint__t *)ptr;
	ptr = ((uint__t *)ptr) + threads;
	nicslu->cluster_end = (uint__t *)ptr;

	/*work space*/
	size = sizeof(void *)*threads;
	ptr = malloc(size);
	if (NULL == ptr)
	{
		return NICSLU_MEMORY_OVERFLOW;
	}
	memset(ptr, 0, size);
	nicslu->workspace_mt1 = (void **)ptr;

	ptr = malloc(size);
	if (NULL == ptr)
	{
		return NICSLU_MEMORY_OVERFLOW;
	}
	memset(ptr, 0, size);
	nicslu->workspace_mt2 = (void **)ptr;

	size = (sizeof(int__t)*4+sizeof(real__t))*n;
	ptr = malloc(size*threads);
	if (NULL == ptr)
	{
		return NICSLU_MEMORY_OVERFLOW;
	}
	for (i=0; i<threads; ++i)
	{
		nicslu->workspace_mt1[i] = ptr;
		ptr = ((byte__t *)ptr) + size;
	}

	size = (sizeof(int__t)*3)*n;
	ptr = malloc(size*threads);
	if (NULL == ptr)
	{
		return NICSLU_MEMORY_OVERFLOW;
	}
	for (i=0; i<threads; ++i)
	{
		nicslu->workspace_mt2[i] = ptr;
		ptr = ((byte__t *)ptr) + size;
	}

	/*create threads*/
	for (i=1; i<threads; ++i)
	{
		if (_CreateThread( \
			_I_NicsLU_ThreadProc, \
			&(((SNicsLUThreadArg *)nicslu->thread_arg)[i]), \
			&(((thread_id__t *)nicslu->thread_id)[i]) \
			) != 0)
		{
			return NICSLU_THREADS_INIT_FAIL;
		}
	}

	/*finish*/
	nicslu->flag[3] = TRUE;

	return NICS_OK;
}

int__t NicsLU_DestroyThreads(SNicsLU *nicslu)
{
	uint__t th;
	bool__t mt;
	uint__t i;
	thread_id__t *thid;
	bool__t *thac;

	if (NULL == nicslu)
	{
		return NICSLU_ARGUMENT_ERROR;
	}

	th = nicslu->cfgi[5];
	mt = nicslu->flag[3];
	thid = (thread_id__t *)(nicslu->thread_id);
	thac = nicslu->thread_active;

	if (mt && th > 1)
	{
		/*set all threads active and wait to exit*/
		nicslu->thread_work = NICSLU_WORK_EXIT;
		for (i=1; i<th; ++i)
		{
			thac[i] = TRUE;
		}
		for (i=1; i<th; ++i)
		{
			_WaitThreadExit(thid[i]);
		}
		for (i=1; i<th; ++i)
		{
			_DestroyThread(thid[i]);
		}
	}
	nicslu->flag[3] = FALSE;
	nicslu->cfgi[5] = 1;
	nicslu->cfgi[7] = 1;
	nicslu->thread_work = NICSLU_WORK_NONE;

	if (nicslu->thread_id != NULL)
	{
		free(nicslu->thread_id);
		nicslu->thread_id = NULL;
	}
	if (nicslu->thread_arg != NULL)
	{
		free(nicslu->thread_arg);
		nicslu->thread_arg = NULL;
	}
	if (nicslu->thread_finish != NULL)
	{
		free(nicslu->thread_finish);
		nicslu->thread_finish = NULL;
	}
	if (nicslu->cluster_start != NULL)
	{
		free(nicslu->cluster_start);
		nicslu->cluster_start = NULL;
	}
	nicslu->cluster_end = NULL;
	if (nicslu->workspace_mt1 != NULL)
	{
		if (nicslu->workspace_mt1[0] != NULL)
		{
			free(nicslu->workspace_mt1[0]);
		}
		free(nicslu->workspace_mt1);
		nicslu->workspace_mt1 = NULL;
	}
	if (nicslu->workspace_mt2 != NULL)
	{
		if (nicslu->workspace_mt2[0] != NULL)
		{
			free(nicslu->workspace_mt2[0]);
		}
		free(nicslu->workspace_mt2);
		nicslu->workspace_mt2 = NULL;
	}
	nicslu->thread_active = NULL;

	return NICS_OK;
}
