#include <stdio.h>
#include "thread.h"

#ifdef _WIN32
/*windows*/

int _CreateThread(thread_proc__t proc, void *arg, thread_id__t *id)
{
	if (proc == NULL || id == NULL) return -1;
	*id = (thread_id__t)_beginthreadex(NULL, 0, proc, arg, 0, NULL);
	return (*id==NULL ? -1 : 0);
}

int _WaitThreadExit(thread_id__t id)
{
	unsigned int rt = WaitForSingleObject(id, INFINITE);
	return (rt==WAIT_FAILED ? -1 : 0);
}

int _DestroyThread(thread_id__t id)
{
	int rt = CloseHandle(id);
	return (rt ? 0 : -1);
}

int _CreateLock(lock__t *lock)
{
	if (lock == NULL) return -1;
	*lock = CreateEvent(NULL, FALSE, TRUE, NULL);
	return (*lock==NULL ? -1 : 0);
}

int _Lock(lock__t *lock)
{
	unsigned int rt;
	if (lock == NULL) return -1;
	rt = WaitForSingleObject(*lock, INFINITE);
	return (rt==WAIT_FAILED ? -1 : 0);
}

int _Unlock(lock__t *lock)
{
	int rt;
	if (lock == NULL) return -1;
	rt = SetEvent(*lock);
	return (rt ? 0 : -1);
}

int _DestroyLock(lock__t *lock)
{
	int rt;
	if (lock == NULL) return -1;
	rt = CloseHandle(*lock);
	return (rt ? 0 : -1);
}

void _Delay(unsigned int ms)
{
	Sleep(ms);
}

int _CreateSemaphore(sem__t *sem, int initval)
{
	if (sem == NULL) return -1;
	*sem = CreateSemaphore(NULL, initval, 2147483647, NULL);
	return (*sem==NULL ? -1 : 0);
}

int _WaitSemaphore(sem__t *sem)
{
	unsigned int rt;
	if (sem == NULL) return -1;
	rt = WaitForSingleObject(*sem, INFINITE);
	return (rt==WAIT_FAILED ? -1 : 0);
}

int _IncreaseSemaphore(sem__t *sem)
{
	int rt;
	if (sem == NULL) return -1;
	rt = ReleaseSemaphore(*sem, 1, NULL);
	return (rt ? 0 : -1);
}

int _DestroySemaphore(sem__t *sem)
{
	int rt;
	if (sem == NULL) return -1;
	rt = CloseHandle(*sem);
	return (rt ? 0 : -1);
}

int _CreateEvent(event__t *ev, int initval)
{
	if (ev == NULL) return -1;
	*ev = CreateEvent(NULL, TRUE, initval>0?TRUE:FALSE, NULL);
	return (*ev==NULL ? -1 : 0);
}

int _WaitEvent(event__t *ev)
{
	unsigned int rt;
	if (ev == NULL) return -1;
	rt = WaitForSingleObject(*ev, INFINITE);
	return (rt==WAIT_FAILED ? -1 : 0);
}

int _SetEvent(event__t *ev)
{
	int rt;
	if (ev == NULL) return -1;
	rt = SetEvent(*ev);
	return (rt ? 0 : -1);
}

int _ResetEvent(event__t *ev)
{
	int rt;
	if (ev == NULL) return -1;
	rt = ResetEvent(*ev);
	return (rt ? 0 : -1);
}

int _DestroyEvent(event__t *ev)
{
	int rt;
	if (ev == NULL) return -1;
	rt = CloseHandle(*ev);
	return (rt ? 0 : -1);
}

#else
/*linux*/

int _CreateThread(thread_proc__t proc, void *arg, thread_id__t *id)
{
	return pthread_create(id, NULL, proc, arg);
}

int _WaitThreadExit(thread_id__t id)
{
	return pthread_join(id, NULL);
}

int _DestroyThread(thread_id__t id)
{
	return 0;
}

int _CreateLock(lock__t *lock)
{
	if (lock == NULL) return -1;
	return pthread_mutex_init(lock, NULL);
}

int _Lock(lock__t *lock)
{
	if (lock == NULL) return -1;
	return pthread_mutex_lock(lock);
}

int _Unlock(lock__t *lock)
{
	if (lock == NULL) return -1;
	return pthread_mutex_unlock(lock);
}

int _DestroyLock(lock__t *lock)
{
	if (lock == NULL) return -1;
	return pthread_mutex_destroy(lock);
}

void _Delay(unsigned int ms)
{
	usleep(ms*1000);
}

int _CreateSemaphore(sem__t *sem, int initval)
{
	if (sem == NULL) return -1;
	return sem_init(sem, 0, initval);
}

int _WaitSemaphore(sem__t *sem)
{
	if (sem == NULL) return -1;
	return sem_wait(sem);
}

int _IncreaseSemaphore(sem__t *sem)
{
	if (sem == NULL) return -1;
	return sem_post(sem);
}

int _DestroySemaphore(sem__t *sem)
{
	if (sem == NULL) return -1;
	return sem_destroy(sem);
}

int _CreateEvent(event__t *ev, int initval)
{
	if (ev == NULL) return -1;
	return sem_init(ev, 0, initval>0?1:0);
}

int _WaitEvent(event__t *ev)
{
	if (ev == NULL) return -1;
	sem_wait(ev);
	return sem_post(ev);
}

int _SetEvent(event__t *ev)
{
	if (ev == NULL) return -1;
	return sem_post(ev);
}

int _ResetEvent(event__t *ev)
{
	if (ev == NULL) return -1;
	sem_trywait(ev);
	return 0;
}

int _DestroyEvent(event__t *ev)
{
	if (ev == NULL) return -1;
	return sem_destroy(ev);
}

#endif

void _SpinWaitInt(volatile int *a)
{
	while (0 == *a);
}

void _SpinWaitShort(volatile short *a)
{
	while (0 == *a);
}

void _SpinWaitChar(volatile char *a)
{
	while (0 == *a);
}

#ifdef _WIN32
void _SpinWaitInt64(volatile __int64 *a)
#else
void _SpinWaitInt64(volatile long long *a)
#endif
{
	while (0 == *a);
}

void _SpinWaitFloat(volatile float *a)
{
	while (0.f == *a);
}

void _SpinWaitDouble(volatile double *a)
{
	while (0. == *a);
}

void _SpinWaitSizeInt(volatile size_t *a)
{
	while (0 == *a);
}

void _SpinBarrier(int s, int n, volatile char *a)
{
	int i;
	volatile char *p;
	for (i=s; i<n; ++i)
	{
		p = &(a[i]);
		while (0 == *p);
	}
}
