/*multi-threaded programming interfaces for both windows and linux*/
/*last modified: apr 14, 2013*/
/*author: Chen, Xiaoming*/
/*system requirements: visual studio compiler for windows and pthread interfaces for linux*/

#ifndef __THREAD__
#define __THREAD__

#include <stddef.h>

#ifdef _WIN32
/*windows*/
#include <process.h>
#define WIN32_LEAN_AND_MEAN
#include <Windows.h>

typedef	unsigned int			(__stdcall *thread_proc__t)(void *);
typedef HANDLE					thread_id__t;
#define THREAD_DECL				unsigned int __stdcall
#define THREAD_DECL_TYPE		unsigned int
#define THREAD_RETURN			0
typedef HANDLE					lock__t;
typedef HANDLE					sem__t;
typedef HANDLE					event__t;

#else
/*linux*/
#include <unistd.h>
#include <pthread.h>
#include <semaphore.h>

typedef void *					(*thread_proc__t)(void *);
typedef pthread_t				thread_id__t;
#define THREAD_DECL				void *
#define THREAD_DECL_TYPE		void *
#define THREAD_RETURN			NULL
typedef pthread_mutex_t			lock__t;
typedef sem_t					sem__t;
typedef sem_t					event__t;

#endif

/*interfaces*/
#ifdef __cplusplus
extern "C" {
#endif

/*if ok, return 0, otherwise fail*/

/*thread*/
int		_CreateThread(thread_proc__t proc, void *arg, thread_id__t *id);
int		_WaitThreadExit(thread_id__t id);
int		_DestroyThread(thread_id__t id);

/*lock, i.e. mutex*/
int		_CreateLock(lock__t *lock);
int		_Lock(lock__t *lock);
int		_Unlock(lock__t *lock);
int		_DestroyLock(lock__t *lock);

/*semaphore*/
int		_CreateSemaphore(sem__t *sem, int initval);
int		_WaitSemaphore(sem__t *sem);
int		_IncreaseSemaphore(sem__t *sem);
int		_DestroySemaphore(sem__t *sem);

/*event*/
int		_CreateEvent(event__t *ev, int initval);
int		_WaitEvent(event__t *ev);
int		_SetEvent(event__t *ev);
int		_ResetEvent(event__t *ev);
int		_DestroyEvent(event__t *ev);

/*delay*/
void	_Delay(unsigned int ms);

/*spin wait, for simple event wait*/
#ifdef _WIN32
void	_SpinWaitInt64(volatile __int64 *);
#else
void	_SpinWaitInt64(volatile long long *);
#endif
void	_SpinWaitInt(volatile int *);
void	_SpinWaitShort(volatile short *);
void	_SpinWaitChar(volatile char *);
void	_SpinWaitFloat(volatile float *);
void	_SpinWaitDouble(volatile double *);
void	_SpinWaitSizeInt(volatile size_t *);
void	_SpinBarrier(int, int, volatile char *);

#ifdef __cplusplus
}
#endif

#endif
