#include "kat.h"
#include "kat_inline.c"
#include "kat_io.h"
#include "kat_fortran.h"
#include "kat_optics.h"
#include "kat_aux.h"
#include "kat_dump.h"
#include "kat_aa.h"
#include "kat_check.h"
#include "kat_knm_mirror.h"
#include "kat_knm_bs.h"
#include "kat_knm_bayer.h"
#include <stdlib.h>
#include <gsl/gsl_sf.h>
#include <gsl/gsl_sf_pow_int.h>


extern init_variables_t init;
extern interferometer_t inter;

double C_n(int n, double x){
    assert(n>=0);
    
    switch(n){
        case 0:
            return 1.0;
        case 1:
            return 2 - x;
        case 2:
            return 6 - 6*x+x*x;
        case 3:
            return 24 - 36*x + 12*x*x - x*x*x;
        case 4:
            return 120 - 240*x + 120*x*x - 20*x*x*x + x*x*x*x;
        default:
            return -(x-2*n)*C_n(n-1, x) - n*(n-1)*C_n(n-2,x);
    }
}

double C_1_n(int n, double x){
    assert(n>=0);
    
    switch(n){
        case 1:
            return -x;
        case 2:
            return -3*x +x*x;
        case 3:
            return -12*x + 8*x*x - x*x*x;
        case 4:
            return -60*x * 60*x*x - 15*x*x*x + x*x*x*x;
        default:
            return -(x-2*n+1)*C_1_n(n-1, x) - n*(n-2)*C_1_n(n-2,x);
    }
}

double B_nmkl(int m, int n, int k, int l, double rhosq){
    if(n==0 && m==0 && k==0 && l==0){
        return 1 - exp(-rhosq);
    }else if(m==0 && n==0 && GSL_IS_EVEN(k) && GSL_IS_EVEN(l)){
        k = k/2;
        l = l/2;
        return pow_neg_1(k+l) * gsl_sf_fact(2*k)* gsl_sf_fact(2*l) / (gsl_sf_fact(k)*gsl_sf_fact(l)*gsl_sf_fact(k+l))
                * rhosq*exp(-rhosq) * C_n(k+l-1, rhosq);
    } else if(m==0 && n==1 && GSL_IS_EVEN(k) && GSL_IS_ODD(l)){
        k = k/2;
        l = (l+1)/2;
        return pow_neg_1(k+l) * gsl_sf_fact(2*k)* gsl_sf_fact(2*l) / (gsl_sf_fact(k)*gsl_sf_fact(l)*gsl_sf_fact(k+l))
                * rhosq*exp(-rhosq) * C_1_n(k+l-1, rhosq);
    } else if(m==1 && n==0 && GSL_IS_ODD(k) && GSL_IS_EVEN(l)){
        k = (k+1)/2;
        l = l/2;
        return pow_neg_1(k+l) * gsl_sf_fact(2*k)* gsl_sf_fact(2*l) / (gsl_sf_fact(k)*gsl_sf_fact(l)*gsl_sf_fact(k+l))
                * rhosq*exp(-rhosq) * C_1_n(k+l-1, rhosq);
    } else
        bug_error("Can't compute this aperture knm factor");
    
    return 0.0;
}

double G_matrix_element(int m, int n, int k, int l, double rsq){
    assert(m>=0);
    assert(n>=0);
    assert(k>=0);
    assert(l>=0);
    
    double c_nm = 1 / sqrt(gsl_sf_pow_int(2.0,n+m) * gsl_sf_fact(n)*gsl_sf_fact(m) );
    double c_lk = 1 / sqrt(gsl_sf_pow_int(2.0,l+k) * gsl_sf_fact(k)*gsl_sf_fact(l) );
    double Gnmkl = 0;
    
    if((n==0 && m==0) || (n==0&&m==1) || (n==1&&m==0))
        Gnmkl = c_nm * c_lk / PI * B_nmkl(n,m,k,l,rsq);
    else if(n > 1)
        Gnmkl = sqrt((l+1)/(double)n) * G_matrix_element(m,n-1,k,l+1,rsq)
                + sqrt(l/(double)n) * G_matrix_element(m,n-1,k,l-1,rsq)
                + sqrt((n-1)/(double)n) * G_matrix_element(m,n-2,k,l,rsq);
    else if(m > 1)    
        Gnmkl = sqrt((k+1)/(double)m) * G_matrix_element(m-1,n,k+1,l,rsq)
                + sqrt(k/(double)m) * G_matrix_element(m-1,n,k-1,l,rsq)
                + sqrt((m-1)/(double)m) * G_matrix_element(m-2,n,k,l,rsq);
    
    return Gnmkl;
        
}

void fill_mirror_knm_aperture_matrix_analytic(mirror_t *mirror, double nr1){
    
    mirror_knm_t *amap = &(mirror->knm_aperture);
    // as we will be calculating aperture knm we need to allocate the memory for it
    // remember that free also needs to be called later
    if (!IS_MIRROR_KNM_ALLOCD(amap))
        bug_error("mirror aperture knm has not been allocated");
    
    warn("EXPERIMENTAL analytic aperture calculation!!");
    
    mirror->knm_aperture.IsIdentities = false;
        
    double rhosq11 = mirror->r_aperture / w_size(mirror->knm_q.qxt1_11, nr1);
    
    int i,j,n,m,k,l;
    
    for(i=0; i<inter.num_fields; i++){
        get_tem_modes_from_field_index(&m,&n,i);
     
        for(j=0; j<inter.num_fields; j++){
            get_tem_modes_from_field_index(&k,&l,j);
            
            mirror->knm_aperture.k11[i][j].re = G_matrix_element(m,n,k,l,rhosq11);
        }
    }    
}