/* 
 * File:   kat_matrix_ccs.h
 * Author: Daniel
 *
 * Created on 12 August 2013, 10:33
 */

#ifndef KAT_MATRIX_CCS_H
#define	KAT_MATRIX_CCS_H

#include "kat.h"

typedef struct pseudocolumn_item {
    int index;
    int type;
    void *ptr;
} pseudocolumn_item_t;

typedef struct pseudocolumn{
    struct pseudocolumn *next;
    
    UT_array *items;
    
    int type;
    void *component;
    
    int value_base_idx; // The index in the values array this column starts
    int nnz; // the number of elements this column takes in the value array
    
} pseudocolumn_t;

typedef struct matrix_ccs {
    double *values;   //!< sparse CCS matrix numerical values (length=2*inter.num_nonzero)
    long *row_idx; //!< sparse CCS matrix row index        (length=inter.num_nonzero)
    long *col_ptr;   //!< sparse CCS matrix column pointer   (length=inter.num_eqns+1)
    long *row_col_count; // stores how many rows per column are present
    
    // variables used to count the number of eqns and nonzeros in the matrix
    int num_eqns;
    int num_nonzeros;
} matrix_ccs_t;

typedef struct ifo_matrix_vars{
    //long *row_col_count; // stores how many rows per column are present
    //double *values;   //!< sparse CCS matrix numerical values (length=2*inter.num_nonzero)
    //long *row_idx; //!< sparse CCS matrix row index        (length=inter.num_nonzero)
    //long *col_ptr;   //!< sparse CCS matrix column pointer   (length=inter.num_eqns+1)
    matrix_ccs_t M;
    double *rhs_values;     //!< ride-hand-side for sparse matrix   (length=2*inter.num_eqns)
    
    pseudocolumn_t *pcol_head, *pcol_curr;
    int pcols_num;
    void *solver_opts; // struct that contains solver specific objects
    
    int type; // STANDARD or SIGNAL_QUANTUM
    
    // frequencies associated with this matrix
    int num_frequencies;
    frequency_t **frequencies;
    
    // modulators are the only component that can couple frequencies. Signal frequencies from
    // other components are handled by including them as sources to the signal sideband matrix
    int ***mod_f_couple_order;  // order of coupling if equivalent mod_does_f_couple is non-zero
    int ***mod_does_f_couple;   // 0 if no coupling, 1 if there is
    int ***mod_f_couple_allocd; // 0 if no memory is to be allocated 1 if it is
    
    // stores the index which a given node occupies in the RHS vector of this matrix
    // maps directly to inter.node_list[i]
    int *node_rhs_idx_1; // rhs index for port 1 of node 
    int *node_rhs_idx_2; // rhs index for port 2 of node 
    int *mirror_rhs_idx;
    int *bs_rhs_idx;
    int *modulator_rhs_idx;
    int *space_rhs_idx;
    int *diode_rhs_idx;
    int *sagnac_rhs_idx;
    int *lens_rhs_idx;
    int *grating_rhs_idx;
    int *output_rhs_idx; /** For a given output index the RHS index of the port this output is attached to is stored */
    int *block_rhs_idx;
    int *feedback_rhs_idx;
} ifo_matrix_vars_t;


void dump_matrix_klu(klu_sparse_var_t *matrix, const char* fname);
void dump_matrix_ccs(matrix_ccs_t *matrix, const char* fname);
void dump_matrix_ccs_labels(ifo_matrix_vars_t *ifo_matrix, const char* fname);

/**
 * Sparse CCS y = M x
 */
void ccs_zgemv(matrix_ccs_t *M, complex_t *x, complex_t *y);

int get_port_rhs_idx(int freq, int mode);
int get_rhs_idx(ifo_matrix_vars_t *M, node_t *node, int port, int freq, int mode);
int get_node_rhs_idx(int port, int freq, int mode, int num_freqs);
int get_comp_rhs_idx2(ifo_matrix_vars_t *matrix, int type, int comp_index);

void check_node_direction();
void alloc_ifo_matrix_ccs(long *bytes, ifo_matrix_vars_t *matrix_ccs);
void alloc_matrix_ccs_modulators(ifo_matrix_vars_t *matrix_ccs);
void build_ccs_matrix(ifo_matrix_vars_t *matrix);
void get_ccs_matrix(ifo_matrix_vars_t *matrix);
void fill_ccs_matrix(ifo_matrix_vars_t *matrix);
void refactor_ccs_matrix(ifo_matrix_vars_t *matrix);
void solve_ccs_matrix(ifo_matrix_vars_t *matrix, double *rhs, bool transpose, bool conjugate);
void solve_ccs_matrix_sparse_input(ifo_matrix_vars_t *matrix, double *rhs, bool transpose, long nblocks, long *blocks);
void clear_ccs_matrix_rhs(ifo_matrix_vars_t *matrix);
void set_rhs_ccs_matrix(long idx, complex_t value, ifo_matrix_vars_t *matrix);
void free_ccs_matrix(ifo_matrix_vars_t *matrix);

#endif	/* KAT_MATRIX_CCS_H */

